import numpy as np

def LR_W(bp, por, ap, sp, wp, clay): 
    """
    Calculate the soil bulk real relative dielectric permittivity using the Lichtenecker and Rother model modified by Wunderlich and return

    This function computes the volumetric water content of a soil 
    mixture using the volumetric mixing model of Lichtenecker and Rother [1], and Wunderlich [2] alpha correction model (LR_W). 
    The model incorporates the bulk dielectric permittivity, bulk and particle densities, and 
    permittivities of air, solid, and water, as well as the soil clay content.

    Parameters
    ----------
    bp : array_like
        Soil bulk real relative dielectric permittivity [-]
    por: array_like
        Soil porosity [m**3/m**3].
    ap : array_like
        Soil air real relative dielectric permittivity phase [-].
    sp : array_like
        Soil solid real relative dielectric permittivity phase [-].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    clay : array_like
        Soil clay content [g/g]*100

    Returns
    -------
    array_like
        Soil volumetric water content [m**3/m**3].

    References
    ----------
    .. [1] Lichtenecker, K., and K. Rother. 1931. 
    Die Herleitung des logarithmischen Mischungsgesetzes aus allgemeinen Prinzipien der staionären Strömung. 
    Phys. Z. 32:255-260.
    .. [2] Tina Wunderlich, Hauke Petersen, Said Attia al Hagrey, Wolfgang Rabbel; 
    Pedophysical Models for Resistivity and Permittivity of Partially Water-Saturated Soils. 
    Vadose Zone Journal 2013;; 12 (4): vzj2013.01.0023. doi: https://doi.org/10.2136/vzj2013.01.0023
    
    Example
    -------
    >>> LR_W(10, 1.3, 2.65, 1, 4, 80, 20)
    0.177
    """
    alpha = -0.46*(clay/100)+0.71
    water = (bp**alpha - (1-por)*sp**alpha - por*ap**alpha) / (wp**alpha - ap**alpha)

    return water


def LR(bp, por, ap, sp, wp, alpha): 
    """
    Calculate the soil volumetric water content using the Lichtenecker and Rother model.

    This function computes the volumetric water content of a soil 
    mixture using the volumetric mixing model of Lichtenecker and Rother [1] (LR). 
    The model incorporates the bulk dielectric permittivity, bulk and particle densities, and 
    permittivities of air, solid, and water.

    Parameters
    ----------
    bp : array_like
        Soil bulk real relative dielectric permittivity [-]
    por: array_like
        Soil porosity [m**3/m**3].
    ap : array_like
        Soil air real relative dielectric permittivity phase [-].
    sp : array_like
        Soil solid real relative dielectric permittivity phase [-].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    alpha : float
        Soil alpha exponent as defined in volumetric mixing theory [-].

    Returns
    -------
    array_like
        Soil volumetric water content [m**3/m**3].

    References
    ----------
    .. [1] Lichtenecker, K., and K. Rother. 1931. 
    Die Herleitung des logarithmischen Mischungsgesetzes aus allgemeinen Prinzipien der staionären Strömung. 
    Phys. Z. 32:255-260.
    
    Example
    -------
    >>> LR(10, 1.3, 2.65, 1, 4, 80, 0.5)
    0.210

    """
    if not isinstance(alpha, np.floating):
        alpha = alpha[0]

    if np.isnan(alpha):
        alpha = 0.5
    water = (bp**alpha - (1-por)*sp**alpha - por*ap**alpha) / (wp**alpha - ap**alpha)

    return water


def LR_MV(bp, por, ap, sp, wp, CEC): 
    """
    Calculate the soil volumetric water content using the Lichtenecker and Rother model modified by Mendoza-Veirana and return

    This function computes the volumetric water content of a soil 
    mixture using the volumetric mixing model of Lichtenecker and Rother [1], and Mendoza Veirana [2] alpha correction model (LR_MV). 
    The model incorporates the bulk dielectric permittivity, bulk and particle densities, and 
    permittivities of air, solid, and water, as well as the soil cation exchange capacity. Reported R2 = 0.94

    Parameters
    ----------
    bp : array_like
        Soil bulk real relative dielectric permittivity [-]
    por: array_like
        Soil porosity [m**3/m**3].
    ap : array_like
        Soil air real relative dielectric permittivity phase [-].
    sp : array_like
        Soil solid real relative dielectric permittivity phase [-].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    CEC : array_like
        Soil cation exchange capacity [meq/100g].

    Returns
    -------
    array_like
        Soil volumetric water content [m**3/m**3].

    References
    ----------
    .. [1] Lichtenecker, K., and K. Rother. 1931. 
    Die Herleitung des logarithmischen Mischungsgesetzes aus allgemeinen Prinzipien der staionären Strömung. 
    Phys. Z. 32:255-260.
    .. [2] Gaston Mendoza Veirana, Jeroen Verhegge, Wim Cornelis, Philippe De Smedt, 
    Soil dielectric permittivity modelling for 50-MHz instrumentation,
    Geoderma, Volume 438, 2023, 116624, ISSN 0016-7061,
    https://doi.org/10.1016/j.geoderma.2023.116624.
    
    Example
    -------
    >>> LR_MV(10, 1.3, 2.65, 1, 4, 80, 20)
    0.078

    """
    alpha = 0.248*np.log(CEC) + 0.366  
    water = (bp**alpha - (1-por)*sp**alpha - por*ap**alpha) / (wp**alpha - ap**alpha)

    return water


import numpy as np
from scipy.constants import pi, epsilon_0

def WunderlichEC(water, ec_init, wat_init, wc, Lw):  
    """
    Calculate the soil bulk real electrical conductivity using the Wunderlich model and return

    This is a effective medium model that uses a differential 
    approach to compute the electrical conductivity based on the initial 
    conditions and the water content [1]. Reported RMSE = 0.0055

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    ec_init : float
        Initial soil bulk real electrical conductivity [S/m].
    wat_init : float
        Initial soil volumetric water content [m**3/m**3].
    wc : array_like
        Soil water real electrical conductivity [S/m].
    Lw : float
        Soil scalar depolarization factor of water aggregates (effective medium theory) [-]

    Returns
    -------
    array_like
        The estimated soil bulk real electrical conductivity [S/m].

    Notes
    -----
    The Wunderlich model is a differential model and the approach used here 
    employs a simple while loop with a step of 0.01 until the differential 
    fraction reaches 1. The method takes into account the initial water content, 
    initial electrical conductivity, and weighting factors to determine the 
    bulk electrical conductivity.

    References
    ----------
    .. [1] Tina Wunderlich, Hauke Petersen, Said Attia al Hagrey, Wolfgang Rabbel; 
    Pedophysical Models for Resistivity and Permittivity of Partially Water-Saturated Soils. 
    Vadose Zone Journal 2013;; 12 (4): vzj2013.01.0023. doi: https://doi.org/10.2136/vzj2013.01.0023

    Example
    -------
    >>> WunderlichEC(0.3, 0.08, 0.05, 0.50, 0.01)
    0.182634

    """
    diff = water - wat_init                                               # Diference utilized just for simplicity
    bulk_ec = ec_init                                                     # Initial permitivity = Epsilon sub 1  
    x = 0                                                                 # Diferentiation from p = 0  
    dx = 0.01                                                             # Diferentiation step

    while x<1:                                                            # Diferentiation until p = 1
        dy = dx*((bulk_ec*diff)/(1-diff+x*diff)) * ((wc-(bulk_ec))/(Lw*wc + (1-Lw)*bulk_ec))
        x=x+dx
        bulk_ec=bulk_ec+dy

    return bulk_ec    


def Fu(water, clay, por, wc, solid_ec, dry_ec, sat_ec, s=1, w=2):
    """
    Calculate the soil bulk real electrical conductivity using the Fu model and return

    This is a volumetric mixing model that takes into account various soil properties 
    such as clay content, porosity, and water content. 
    It was exhaustively validated using several soil samples [1]. Reported R2 = 0.98

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    clay : array_like
        Soil clay content [g/g]*100.
    por: array_like
        Soil porosity [m**3/m**3].
    wc : array_like
        Soil water real electrical conductivity [S/m].
    solid_ec : array_like
        Soil solid real electrical conductivity [S/m].
    dry_ec : array_like
        Soil bulk real electrical conductivity at zero water content [S/m].
    sat_ec : array_like
        Soil bulk real electrical conductivity at saturation water content [S/m].
    s : float, optional
        Phase exponent of the solid, default is 1.
    w : float, optional
        Phase exponent of the water, default is 2.

    Returns
    -------
    array_like
        The estimated bulk electrical conductivity [S/m].

    Notes
    -----
    The method uses default values for s and w, which are 1 and 2 respectively, 
    but can be modified if necessary. Three different forms of the model are used 
    depending on the soil data availability. The soil electrical conductivity of solid surfaces 
    is calculated as in [1] using the formula of Doussan and Ruy (2009) [2]

    References
    ----------
    .. [1] Yongwei Fu, Robert Horton, Tusheng Ren, J.L. Heitman,
    A general form of Archie's model for estimating bulk soil electrical conductivity,
    Journal of Hydrology, Volume 597, 2021, 126160, ISSN 0022-1694, https://doi.org/10.1016/j.jhydrol.2021.126160.
    .. [2] Doussan, C., and Ruy, S. (2009), 
    Prediction of unsaturated soil hydraulic conductivity with electrical conductivity, 
    Water Resour. Res., 45, W10408, doi:10.1029/2008WR007309.

    Example
    -------
    >>> Fu(0.3, 30, 1.3, 2.65, 0.3, 0, np.nan, np.nan)
    0.072626

    """
    d = 0.6539
    e = 0.0183
    surf_ec = (d*clay/(100-clay))+e # Soil electrical conductivity of solid surfaces

    if np.isnan(dry_ec) & np.isnan(sat_ec):
        bulk_ec = solid_ec*(1-por)**s + (water**(w-1))*(por*surf_ec) + wc*water**w

    elif ~(np.isnan(dry_ec)) & ~(np.isnan(sat_ec)):
        bulk_ec = dry_ec + ((dry_ec-sat_ec)/(por**w) - surf_ec)*water**w + (water**(w-1))*(por*surf_ec)

    elif ~(np.isnan(dry_ec)) & np.isnan(sat_ec):
        sat_ec = dry_ec + (wc+surf_ec)*por**w
        bulk_ec = dry_ec + ((dry_ec-sat_ec)/(por**w) - surf_ec)*water**w + (water**(w-1))*(por*surf_ec)

    return bulk_ec


def LongmireSmithEC(bulk_ec_dc, frequency_ec):
    """
    Calculate the soil bulk real electrical conductivity using the Longmire-Smith model and return

    This is a semiempirical model that calculates the soil bulk real electrical conductivity at different
    electromagnetic frequencies [1].

    Parameters
    ----------
    bulk_ec_dc : array_like
        Soil bulk real direct current electrical conductivity [S/m].
    frequency_ec : array_like
        Frequency of electric conductivity measurement [Hz].

    Returns
    -------
    array_like
        Soil bulk real electrical conductivity [S/m].

    Notes
    -----
    The Longmire-Smith equation uses a set of coefficients to account for the 
    frequency-dependent dielectric dispersion. If all values in the `bulk_ec_dc` 
    array are zero, the function returns 0.

    Global Variables Used
    ---------------------
    epsilon_0 : float
        The vacuum permittivity constant.

    References
    ----------
    .. [1] K. S. Smith and C. L. Longmire, “A universal impedance for soils,” 
    Defense Nuclear Agency, Alexandria, VA, USA, Topical 
    Report for Period Jul. 1 1975-Sep. 30 1975, 1975.

    Example
    -------
    >>> LongmireSmithEC(np.array([0.05, 0.10]), 130)
    array([0.05153802, 0.10245936])

    """
    if (bulk_ec_dc == 0).all():
        return 0
    
    else: 
        a = [3.4e6, 2.74e5, 2.58e4, 3.38e3, 5.26e2, 1.33e2, 2.72e1, 1.25e1, 4.8, 2.17, 9.8e-1, 3.92e-1, 1.73e-1]
        f = (125*bulk_ec_dc)**0.8312
        bulk_eci_ = []
        
        for i in range(len(a)):
            F_ = f*(10**i)
            bulk_eci = 2*pi*epsilon_0*(a[i]*F_*(frequency_ec/F_)**2/(1+(frequency_ec/F_)**2))                     
            bulk_eci_.append(bulk_eci)

        bulk_ec = bulk_ec_dc + sum(bulk_eci_)
        return bulk_ec
    
    
def Rhoades(water, wc, s_ec, E, F):
    """
    Calculate the soil bulk real electrical conductivity using the Rhoades model and return

    This function estimates the bulk electrical conductivity of a soil-water mixture 
    using the Rhoades equation [1]. The model combines the contributions of the water's 
    electrical conductivity and the soil's electrical conductivity, adjusted by 
    weighting factors based on the water content.

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    wc : array_like
        Soil water real electrical conductivity [S/m].
    s_ec : array_like
        Soil bulk real surface electrical conductivity [S/m].
    E : float
        Weighting factor for the quadratic term of water content in the Rhoades equation.
    F : float
        Weighting factor for the linear term of water content in the Rhoades equation.

    Returns
    -------
    array_like
        Soil bulk real electrical conductivity [S/m].

    References
    ----------
    .. [1] Rhoades, J.D., Raats, P.A.C. and Prather, R.J. (1976), 
    Effects of Liquid-phase Electrical Conductivity, Water Content, 
    and Surface Conductivity on Bulk Soil Electrical Conductivity. 
    Soil Science Society of America Journal, 40: 651-655. 
    https://doi.org/10.2136/sssaj1976.03615995004000050017x

    Example
    -------
    >>> Rhoades(0.3, 0.5, 0.001, 1, 0.5)
    0.121

    """
    bulk_ec = wc*(E*(water**2)+F*water) + s_ec
    return bulk_ec


def SheetsHendrickx(bulk_ec, temperature):
    """
    Calculate the soil bulk real electrical conductivity using the Sheets-Hendricks model and return

    This function adjusts the apparent electrical conductivity (ECa) of soil to a standard temperature of 298.15 K (25°C). The adjustment is based on the Sheets-Hendricks model.

    Parameters
    ----------
    bulk_ec : array-like
        Soil bulk real electrical conductivity [S/m]
    temperature : array-like
        Soil bulk temperature [K]

    Returns
    -------
    array_like
        Soil bulk real electrical conductivity temperature corrected [S/m]

    Notes
    -----
    The Sheets-Hendricks model applies a temperature correction factor to adjust the apparent electrical conductivity to a standard temperature of 25°C. This correction is particularly important in precision agriculture and soil science studies where temperature fluctuations can significantly affect conductivity measurements.

    Example
    -------
    >>> SheetsHendrickxEC(np.array([1.2, 2.5]), 20)
    array([0.13352103, 0.27816881])
    """
    temp_c = temperature - 273.15
    ft = 0.447+1.4034*np.exp(-temp_c/26.815) # Temperature conversion factor
    bulk_ec_tc = bulk_ec*ft
    return bulk_ec_tc


import numpy as np

from .bulk_ec import BulkEC

from pedophysics.pedophysical_models.bulk_ec import SheetsHendrickx

def BulkECTC(soil):
    """
    Calculate missing values of soil.df.bulk_ec_tc and return

    This function checks if any values of `bulk_ec_tc` (temperature-corrected bulk electrical conductivity)
    in `soil.df` are missing. If so, it updates these values based on the soil's temperature. For temperatures
    exactly at 298.15K, `bulk_ec_tc` is set equal to `bulk_ec` without correction. For other temperatures,
    `bulk_ec_tc` is calculated using the SheetsHendrickx function. Corresponding entries in `soil.info['bulk_ec_tc']`
    are annotated to indicate whether values were directly taken from `bulk_ec` or calculated using the
    SheetsHendrickx function.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            Data Frame containing quantitative information of soil attributes for each state, including 
            `bulk_ec`, `bulk_ec_tc`, and `temperature`.
        - info: DataFrame
            Dictionary containing descriptive information about how each attribute was calculated.
        - n_states: int
            Number of soil states.

    Returns
    -------
    np.ndarray
        soil.df.bulk_ec_tc.values: an array of updated Soil bulk real electrical conductivity temperature corrected (298.15 K) values

    External functions
    --------
    BulkEC : Calculate missing values of soil.df.bulk_ec and return
    SheetsHendrickx : Calculate the soil bulk real electrical conductivity using the Sheets-Hendricks model and return

    Notes
    -----
    - If `bulk_ec_tc` is NaN and temperature is 298.15K, `bulk_ec_tc` is set equal to `bulk_ec` with an annotation
      indicating no temperature correction was applied.
    - If `bulk_ec_tc` is NaN and temperature is not 298.15K, `bulk_ec_tc` is calculated using the SheetsHendrickx
      function, with an annotation indicating it was temperature-corrected.
    - The function updates the soil object in-place and returns an array of the updated `bulk_ec_tc` values.

    """
    # Check if any value of bulk_perm_inf is missing
    if (np.isnan(soil.df.bulk_ec_tc)).any():
        BulkEC(soil)

        # Check for missing values
        missing_bulk_ec_tc_before = soil.df['bulk_ec_tc'].isna()

        soil.df['bulk_ec_tc'] = [soil.df.bulk_ec.values[x] if np.isnan(soil.df.bulk_ec_tc[x]) and soil.df.temperature[x] == 298.15 
                                 else soil.df.bulk_ec_tc[x] for x in range(soil.n_states)] # There is no temperaturre correction
        
        missing_bulk_ec_tc_after = soil.df['bulk_ec_tc'].isna()

        soil.info['bulk_ec_tc'] = [str(soil.info.bulk_ec_tc[x]) + (
                "--> Equal to soil.df.bulk_ec because temperature = 298.15"
                if missing_bulk_ec_tc_before[x] and not missing_bulk_ec_tc_after[x]
                else "--> Provide bulk_ec_tc; otherwise, bulk_ec, and temperature"
                if missing_bulk_ec_tc_before[x] and missing_bulk_ec_tc_after[x]
                else "")
            if missing_bulk_ec_tc_before[x]
            else soil.info.bulk_ec_tc[x]
            for x in range(soil.n_states)]


        # Check for missing values
        missing_bulk_ec_tc_before = soil.df['bulk_ec_tc'].isna()

        soil.df['bulk_ec_tc'] = [SheetsHendrickx(soil.df.bulk_ec.values[x], soil.df.temperature.values[x]) 
                                    if np.isnan(soil.df.bulk_ec_tc[x]) and soil.df.temperature[x] != 298.15
                                    else soil.df.bulk_ec_tc[x] for x in range(soil.n_states)]
        
        missing_bulk_ec_tc_after = soil.df['bulk_ec_tc'].isna()
                
        soil.info['bulk_ec_tc'] = [str(soil.info.bulk_ec_tc[x]) + (
                "--> Calculated using SheetsHendrickx function in predict.bulk_ec_tc.BulkECTC"
                if missing_bulk_ec_tc_before[x] and not missing_bulk_ec_tc_after[x]
                else "--> Provide bulk_ec_tc; otherwise, bulk_ec"
                if missing_bulk_ec_tc_before[x] and missing_bulk_ec_tc_after[x]
                else "")
            if missing_bulk_ec_tc_before[x]
            else soil.info.bulk_ec_tc[x]
            for x in range(soil.n_states)]

    return soil.df.bulk_ec_tc.values



import numpy as np

def WunderlichP(water, perm_init, wat_init, wp, Lw): 
    """
    Calculate the soil bulk real relative dielectric permittivity using the Wunderlich model and return

    This is a effective medium model that uses a differential 
    approach to compute the dielectric permittivity based on the initial 
    conditions and the water content [1]. Reported RMSE = 0.009

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    perm_init : float
        Initial soil bulk real relative dielectric permittivity [-].
    wat_init : float
        Initial soil volumetric water content [m**3/m**3].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    Lw : float
        Soil scalar depolarization factor of water aggregates (effective medium theory) [-]

    Returns
    -------
    array_like
        The estimated soil bulk real relative dielectric permittivity [-]

    Notes
    -----
    The Wunderlich model is a differential model and the approach used here 
    employs a simple while loop with a step of 0.01 until the differential 
    fraction reaches 1. The method takes into account the initial water content, 
    initial relative dielectric permittivity, and weighting factors to determine the 
    bulk real relative dielectric permittivity.

    References
    ----------
    .. [1] Tina Wunderlich, Hauke Petersen, Said Attia al Hagrey, Wolfgang Rabbel; 
    Pedophysical Models for Resistivity and Permittivity of Partially Water-Saturated Soils. 
    Vadose Zone Journal 2013;; 12 (4): vzj2013.01.0023. doi: https://doi.org/10.2136/vzj2013.01.0023

    Example
    -------
    >>> WunderlichP(0.3, 7, 0.05, 80, 0.01)
    24.591

    """
    diff = water - wat_init                                        # Diference utilized just for simplicity
    bulk_perm = perm_init                                          # Initial permitivity = epsilon sub 1  
    x = 0.001                                                      # Diferentiation from p = 0  
    dx = 0.01                                                      # Diferentiation step
                                                                   # Diferentiation until p = 1
    while x<1:                                                    
        dy = ((bulk_perm*diff)/(1-diff+x*diff)) * ((wp-bulk_perm)/(Lw*wp+(1-Lw)*bulk_perm))
        x=x+dx
        bulk_perm = bulk_perm+dy*dx
        
    return bulk_perm


def LR_MV(water, por, ap, sp, wp, CEC): 
    """
    Calculate the soil bulk real relative dielectric permittivity using the Lichtenecker-Rother model modified by Mendoza-Veirana and return

    This function computes the bulk real relative dielectric permittivity of a soil 
    mixture using the volumetric mixing model of Lichtenecker and Rother [1], and Mendoza Veirana [2] alpha correction model (LR_MV). 
    The model incorporates the water content, bulk and particle densities, and 
    permittivities of air, solid, and water, as well as the soil cation exchange capacity. Reported R2 = 0.94

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    por: array_like
        Soil porosity [m**3/m**3].
    ap : array_like
        Soil air real relative dielectric permittivity phase [-].
    sp : array_like
        Soil solid real relative dielectric permittivity phase [-].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    CEC : array_like
        Soil cation exchange capacity [meq/100g].

    Returns
    -------
    array_like
        The estimated soil bulk real relative dielectric permittivity [-]

    References
    ----------
    .. [1] Lichtenecker, K., and K. Rother. 1931. 
    Die Herleitung des logarithmischen Mischungsgesetzes aus allgemeinen Prinzipien der staionären Strömung. 
    Phys. Z. 32:255-260.
    .. [2] Gaston Mendoza Veirana, Jeroen Verhegge, Wim Cornelis, Philippe De Smedt, 
    Soil dielectric permittivity modelling for 50-MHz instrumentation,
    Geoderma, Volume 438, 2023, 116624, ISSN 0016-7061,
    https://doi.org/10.1016/j.geoderma.2023.116624.
    
    Example
    -------
    >>> LR_MV(0.3, 1.3, 2.65, 1, 4, 80, 20)
    28.577  

    """
    alpha = 0.248*np.log(CEC) + 0.366
    bulk_perm = ( water*wp**alpha + (1-por)*sp**alpha + (por-water)*ap**(alpha))**(1/alpha)

    return bulk_perm


def LR(water, por, ap, sp, wp, alpha): 
    """
    Calculate the soil bulk real relative dielectric permittivity using the Lichtenecker and Rother model and return

    This function computes the bulk real relative dielectric permittivity of a soil 
    mixture using the volumetric mixing model of Lichtenecker and Rother [1] (LR). 
    The model incorporates the water content, bulk and particle densities, and 
    permittivities of air, solid, and water.

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    por: array_like
        Soil porosity [m**3/m**3].
    ap : array_like
        Soil air real relative dielectric permittivity phase [-].
    sp : array_like
        Soil solid real relative dielectric permittivity phase [-].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    alpha : float
        Soil alpha exponent as defined in volumetric mixing theory [-].

    Returns
    -------
    array_like
        The estimated soil bulk real relative dielectric permittivity [-]

    References
    ----------
    .. [1] Lichtenecker, K., and K. Rother. 1931. 
    Die Herleitung des logarithmischen Mischungsgesetzes aus allgemeinen Prinzipien der staionären Strömung. 
    Phys. Z. 32:255-260.
    
    Example
    -------
    >>> LR(0.3, 1.3, 2.65, 1, 4, 80, 0.5)
    15.006

    """
    if not isinstance(alpha, np.floating):
        alpha = alpha[0]
    bulk_perm = ( water*wp**alpha + (1-por)*sp**alpha + (por-water)*ap**(alpha))**(1/alpha)

    return bulk_perm


def LR_W(water, por, ap, sp, wp, clay): 
    """
    Calculate the soil bulk real relative dielectric permittivity using the Lichtenecker and Rother model modified by Wunderlich and return

    This function computes the bulk real relative dielectric permittivity of a soil 
    mixture using the volumetric mixing model of Lichtenecker and Rother [1], and Wunderlich [2] alpha correction model (LR_W). 
    The model incorporates the water content, bulk and particle densities, and 
    permittivities of air, solid, and water, as well as the soil clay content.

    Parameters
    ----------
    water : array_like
        Soil volumetric water content [m**3/m**3].
    por: array_like
        Soil porosity [m**3/m**3].
    ap : array_like
        Soil air real relative dielectric permittivity phase [-].
    sp : array_like
        Soil solid real relative dielectric permittivity phase [-].
    wp : array_like
        Soil water phase real dielectric permittivity [-].
    clay : array_like
        Soil clay content [g/g]*100

    Returns
    -------
    array_like
        The estimated soil bulk real relative dielectric permittivity [-]

    References
    ----------
    .. [1] Lichtenecker, K., and K. Rother. 1931. 
    Die Herleitung des logarithmischen Mischungsgesetzes aus allgemeinen Prinzipien der staionären Strömung. 
    Phys. Z. 32:255-260.
    .. [2] Tina Wunderlich, Hauke Petersen, Said Attia al Hagrey, Wolfgang Rabbel; 
    Pedophysical Models for Resistivity and Permittivity of Partially Water-Saturated Soils. 
    Vadose Zone Journal 2013;; 12 (4): vzj2013.01.0023. doi: https://doi.org/10.2136/vzj2013.01.0023
    
    Example
    -------
    >>> LR_W(0.3, 1.3, 2.65, 1, 4, 80, 20)
    17.505 

    """
    alpha = -0.46*(clay/100)+0.71
    bulk_perm = ( water*wp**alpha + (1-por)*sp**alpha + (por-water)*ap**(alpha))**(1/alpha)

    return bulk_perm


def LongmireSmithP(bulk_ec_dc, bulk_perm_inf, frequency_perm):
    """
    Calculate the soil bulk real relative dielectric permittivity using the Longmire-Smith model and return

    This is a semiempirical model that calculates the soil bulk real relative dielectric permittivity at different
    electromagnetic frequencies [1].

    Parameters
    ----------
    bulk_ec_dc : array_like
        Soil bulk real direct current electrical conductivity [-].
    bulk_perm_inf : array_like
        Soil bulk real relative permittivity at infinite frequency [-].
    frequency_perm : array_like
        Frequency of dielectric permittivity measurement [Hz].

    Returns
    -------
    array_like
        Soil bulk real relative dielectric permittivity [-].

    Notes
    -----
    The Longmire-Smith equation uses a set of coefficients to account for the 
    frequency-dependent dielectric dispersion. If all values in the `bulk_ec_dc` 
    array are zero, the function returns 'bulk_perm_inf'.

    Global Variables Used
    ---------------------
    epsilon_0 : float
        The vacuum permittivity constant.

    References
    ----------
    .. [1] K. S. Smith and C. L. Longmire, “A universal impedance for soils,” 
    Defense Nuclear Agency, Alexandria, VA, USA, Topical 
    Report for Period Jul. 1 1975-Sep. 30 1975, 1975.

    Example
    -------
    >>> LongmireSmithP(0.1, 5, 50e6)
    23.328

    """
    
    if (bulk_ec_dc == 0).all():
        return bulk_perm_inf
    
    a = [3.4e6, 2.74e5, 2.58e4, 3.38e3, 5.26e2, 1.33e2, 2.72e1, 1.25e1, 4.8, 2.17, 9.8e-1, 3.92e-1, 1.73e-1]
    f = (125*bulk_ec_dc)**0.8312
    bulk_permi_ = []

    for i in range(len(a)):
        F_ = f*(10**i)
        bulk_permi = a[i]/(1+(frequency_perm/F_)**2)
        bulk_permi_.append(bulk_permi)
    bulk_perm = bulk_perm_inf + sum(bulk_permi_)

    return bulk_perm


def Hilhorst(bulk_ec, water_ec, water_perm, offset_perm):
    """
    Calculate the soil bulk real relative dielectric permittivity using the Hilhorst model and return

    This function calculates the soil bulk real relative dielectric permittivity of a 
    soil-water mixture based on Hilhorst's model. The relation 
    connects the bulk electrical conductivity of the mixture with the permittivity 
    of the water phase and an offset for the permittivity.

    Parameters
    ----------
    bulk_ec : array_like
        Soil bulk real relative dielectric permittivity [-].
    water_ec : array_like
        Soil water real electrical conductivity [S/m].
    water_perm : array_like
        Soil water phase real dielectric permittivity [-]. 
    offset_perm : array_like
        Soil bulk real relative dielectric permittivity when soil bulk real electrical conductivity is zero [-].

    Returns
    -------
    array_like
        Soil bulk real relative dielectric permittivity [-].

    References
    ----------
    .. [1] Hilhorst, M.A. (2000), A Pore Water Conductivity Sensor. 
    Soil Sci. Soc. Am. J., 64: 1922-1925. https://doi.org/10.2136/sssaj2000.6461922x   

    Example
    -------
    >>> Hilhorst(0.05, 0.5, 80, 4)
    12.0

    """
    bulk_perm = bulk_ec*water_perm/water_ec + offset_perm

    return bulk_perm


import pandas as pd
import numpy as np

def Inst2FreqP(soil):
    """
    Set missing values of soil.df.frequency_perm and return
    
    This function iterates over each state in the 'soil' object. It checks the instrument type associated with the soil ('GPR', 'TDR', or 'HydraProbe'). 
    If the current 'frequency_perm' value is NaN, the function sets a default frequency value specific to the instrument type 
    (1e9 Hz for 'GPR', 200e6 Hz for 'TDR', and 50e6 Hz for 'HydraProbe') in both the 'info' attribute and the 'df' DataFrame. 
    If the 'frequency_perm' value is already set or the instrument type does not match any specified case, the function retains the existing value.

    Parameters
    ----------
    soil : object
        A custom soil object that contains:

        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state

    Returns
    -------
    np.ndarray
        Array containing the updated frequency of dielectric permittivity measurement values

    Notes
    -----
    This function modifies the soil object in-place by updating the `df` and `info` dataframes.
    """

    soil.info['frequency_perm'] = ["Set as 1e9 Hz because soil.instrument == GPR" if ((soil.instrument == 'GPR') & np.isnan(soil.df.frequency_perm[x])) or (soil.info.frequency_perm[x] == "Set as 1e9 because soil.instrument == GPR")
                                   else soil.info.frequency_perm[x] for x in range(soil.n_states)]

    soil.df['frequency_perm'] = [1e9 if (soil.instrument == 'GPR') & np.isnan(soil.df.frequency_perm[x]) else soil.df.frequency_perm[x] for x in range(soil.n_states)]

    soil.info['frequency_perm'] = ["Set as 200e6 Hz because soil.instrument == TDR" if ((soil.instrument == 'TDR') & np.isnan(soil.df.frequency_perm[x])) or (soil.info.frequency_perm[x] == "Set as 200e6 because soil.instrument == TDR")
                                   else soil.info.frequency_perm[x] for x in range(soil.n_states)]
    
    soil.df['frequency_perm'] = [200e6 if (soil.instrument == 'TDR') & np.isnan(soil.df.frequency_perm[x]) else soil.df.frequency_perm[x] for x in range(soil.n_states)]

    soil.info['frequency_perm'] = ["Set as 50e6 Hz because soil.instrument == HydraProbe" if ((soil.instrument == 'HydraProbe') & np.isnan(soil.df.frequency_perm[x])) or (soil.info.frequency_perm[x] == "Set as 50e6 because soil.instrument == HydraProbe")
                                   else soil.info.frequency_perm[x] for x in range(soil.n_states)]
    
    soil.df['frequency_perm'] = [50e6 if (soil.instrument == 'HydraProbe') & np.isnan(soil.df.frequency_perm[x]) else soil.df.frequency_perm[x] for x in range(soil.n_states)]

    return soil.df.frequency_perm


def Inst2FreqEC(soil):
    """
    Set missing values of soil.df.frequency_ec and return

    This function iterates over each state in the 'soil' object, checking the type of instrument associated with the 'soil' object ('EMI Dualem' or 'EMI EM38-DD'). 
    If the current 'frequency_ec' value is NaN, the function sets a default frequency value (9e3 Hz for 'EMI Dualem' and 16e3 Hz for 'EMI EM38-DD') 
    in both the 'info' attribute and the 'df' dataframe. 
    If the 'frequency_ec' value is already set or the instrument type does not match, the function retains the existing value.

    Parameters
    ----------
    soil : object
        A custom soil object that contains:

        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state

    Returns
    -------
    np.ndarray
        Array containing the updated frequency of electric conductivity measurement values

    Notes
    -----
    This function modifies the soil object in-place by updating the `df` and `info` dataframes.
    """
    
    soil.info['frequency_ec'] = ["Set as 9e3 Hz because soil.instrument == EMI Dualem" if (soil.instrument == 'EMI Dualem') & np.isnan(soil.df.frequency_ec[x]) or (soil.info.frequency_ec[x] == "Set as 9e3 because soil.instrument == EMI Dualem")
                                   else soil.info.frequency_ec[x] for x in range(soil.n_states)]
    
    soil.df['frequency_ec'] = [9e3 if (soil.instrument == 'EMI Dualem') & np.isnan(soil.df.frequency_ec[x]) else soil.df.frequency_ec[x] for x in range(soil.n_states)]

    soil.info['frequency_ec'] = ["Set as 16e3 Hz because soil.instrument == EMI EM38-DD" if ((soil.instrument == 'EMI EM38-DD') & np.isnan(soil.df.frequency_ec[x])) or (soil.info.frequency_ec[x] == "Set as 16e3 because soil.instrument == EMI EM38-DD")
                                   else soil.info.frequency_ec[x] for x in range(soil.n_states)]
    
    soil.df['frequency_ec'] = [16e3 if (soil.instrument == 'EMI EM38-DD') & np.isnan(soil.df.frequency_ec[x]) else soil.df.frequency_ec[x] for x in range(soil.n_states)]

    return soil.df.frequency_ec


import numpy as np

def SolidEC(soil):
    """
    Set missing values of soil.df.solid_ec and return

    If any value of the solid_ec attribute is missing (NaN), 
    it will be set to a default value of 0. Corresponding information 
    in the soil's info dataframe will be updated accordingly.

    Parameters
    ----------
    soil : object
        A custom soil object that contains:

        - solid_ec : array-like
            Soil solid real electrical conductivity [S/m]
        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state
        - info : DataFrame
            Data Frame containing descriptive information about how each array-like attribute was determined or modified.
        - n_states : int
            Number of states or records in the dataframe.

    Returns
    -------
    np.ndarray
        soil.df.solid_ec.values: an array of updated soil solid real electrical conductivity values

    Example
    -------
    >>> sample = Soil()
    >>> sample.df.solid_ec
    0   NaN
    Name: solid_ec, dtype: float64
    >>> SolidEC(sample)
    >>> sample.df.solid_ec
    0    0
    Name: solid_ec, dtype: float64
    """

    # Check if any value of solid_ec is missing
    if (np.isnan(soil.df.solid_ec)).any():

        soil.info['solid_ec'] = ["Set as zero by default" if np.isnan(soil.df.solid_ec[x]) or soil.info.solid_ec[x] == "Set as zero by default"
                                 else soil.info.solid_ec[x] for x in range(soil.n_states)]

        soil.df['solid_ec'] = [0 if np.isnan(soil.df.solid_ec[x]) else soil.df.solid_ec[x] for x in range(soil.n_states)]

    return soil.df.solid_ec.values


import numpy as np
from pedophysics import instruments

def FrequencyEC(soil): 
    """
    Return and set missing values of the soil.df.frequency_ec attribute.

    If any value of the frequency_ec attribute is missing (NaN), it will be set to a default value of 0 Hz (direct current). 
    Corresponding information in the soil's info dataframe will be updated accordingly.

    Parameters
    ----------
    soil : object
        A custom soil object that contains:
        - frequency_ec : array-like
            Frequency of electric conductivity measurement [Hz]
        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state.
        - info : DataFrame
            Data Frame containing descriptive information about how each array-like attribute was calculated. includes: frequency_ec
        - n_states : int
            Number of states or records in the dataframe.

    Returns
    -------
    np.ndarray
        soil.df.frequency_ec.values: an array of updated frequency of electric conductivity measurement values.

    External functions
    ------------------
    Inst2FreqEC : Function to calculate missing frequency_ec attribute based on soil.instrument.

    Example
    -------
    >>> sample = Soil()
    >>> sample.df.frequency_ec
    0   NaN
    Name: frequency_ec, dtype: float64
    >>> FrequencyEC(sample)
    >>> sample.df.frequency_ec
    0    0
    Name: frequency_ec, dtype: float64
    """

    # Check if any value of frequency_ec is missing
    if (np.isnan(soil.df.frequency_ec)).any():
        instruments.Inst2FreqEC(soil)

        soil.info['frequency_ec'] = ["Set as 0 Hz (direct current) by default" if np.isnan(soil.df.frequency_ec[x]) or soil.info.frequency_ec[x] == "Set as 0 Hz (direct current) by default"
                                     else soil.info.frequency_ec[x] for x in range(soil.n_states)]
        
        soil.df['frequency_ec'] = [0 if np.isnan(soil.df.frequency_ec[x]) else soil.df.frequency_ec[x] for x in range(soil.n_states)]

    return soil.df.frequency_ec.values


import numpy as np
from pedophysics import instruments

def FrequencyPerm(soil): 
    """
    Set missing values of soil.df.frequency_perm and return 

    If any value of the frequency_perm attribute is missing (NaN), 
    it will be computed using the `Inst2FreqP` function from the `instruments` module. 

    Parameters
    ----------
    soil : object
        A custom soil object that contains:

        - frequency_perm : array-like
            Frequency of dielectric permittivity measurement [Hz]
        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state. Includes: frequency_perm

    Returns
    -------
    np.ndarray
        soil.df.frequency_perm.values: an array of updated frequency of dielectric permittivity measurement values

    Notes
    -----
    This function modifies the soil object in-place, updating the `df` dataframe 
    if necessary.

    External functions
    --------
    Inst2FreqP : Set missing values of soil.df.frequency_perm and return

    Example
    -------
    >>> sample = Soil(instrument = 'GPR')
    >>> sample.df.frequency_perm
    0   NaN
    Name: frequency_perm, dtype: float64
    >>> FrequencyPerm(sample)
    >>> sample.df.frequency_perm
    0    1e9
    Name: frequency_perm, dtype: float64
    """

    # Check if any value of frequency_perm is missing
    if (np.isnan(soil.df.frequency_perm)).any(): 
        instruments.Inst2FreqP(soil)

    return soil.df.frequency_perm.values


from .air_perm import *
from .bulk_ec import *
from .bulk_ec_tc import *
from .bulk_ec_dc import *
from .bulk_ec_dc_tc import *
from .bulk_perm_inf import *
from .bulk_perm import *
from .frequency_ec import *
from .frequency_perm import *
from .particle_density import *
from .porosity import *
from .solid_ec import *
from .solid_perm import *
from .temperature import *
from .texture import *
from .water_ec import *
from .water_perm import *
from .water import *
from .water_from_ec import *
from .water_from_perm import *
from .salinity import *


import numpy as np
from scipy.optimize import minimize

from .frequency_ec import *
from .solid_ec import *
from .temperature import *
from .bulk_ec_dc import non_dc_to_dc
from .texture import Texture
from .porosity import Porosity

from pedophysics.pedophysical_models.bulk_ec import Fu, SheetsHendrickx, WunderlichEC
from pedophysics.utils.stats import R2_score


def BulkECDCTC(soil):
    """ 
    Compute missing values of soil.df.bulk_ec_dc_tc and return

    This function initiates a series of steps to estimate missing `bulk_ec_dc_tc` values in the soil DataFrame.
    It starts by updating the soil's frequency and temperature data using `FrequencyEC` and `Temperature` functions,
    respectively, followed by a shift to `bulk_ec_dc_tc` values through `shift_to_bulk_ec_dc_tc`. If the conditions
    for a fitting routine are met, indicated by at least three states having non-NaN values for both water content
    and `bulk_ec_dc_tc`, the `fitting` function is called to further refine these estimates. Conversely, if there
    are states with non-NaN water content but missing `bulk_ec_dc_tc` values, the `non_fitting` routine is invoked
    to estimate these missing values.

    Parameters
    ----------
    soil : object
        A custom soil object containing:
        - df : DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state.
            Includes: bulk_ec, bulk_ec_dc_tc.
        - n_states : int
            Number of soil states.

    Returns
    -------
    numpy.ndarray
        soil.df.bulk_ec_dc_tc.values: array containing updated soil bulk real electrical conductivity direct current (0 Hz) temperature corrected (298.15 K) values

    Notes
    -----
    - The function operates in-place, modifying the soil object's DataFrame directly.
    - It relies on both fitting and non-fitting routines to estimate missing `bulk_ec_dc_tc` values, depending
      on the availability and condition of the soil data.
    - The estimation process is contingent upon the presence of adequate non-NaN data points for both water content
      and `bulk_ec_dc_tc` across the soil states.

    External Functions
    ------------------
    - FrequencyEC : Set missing values of soil.df.frequency_ec and return 
    - Temperature : Set missing values of soil.df.temperature and return 
    - shift_to_bulk_ec_dc_tc : Compute missing values of soil.df.bulk_ec_dc_tc based on soil.df.bulk_ec or soil.df.bulk_ec_dc
    - fitting : Calculate missing values of soil.df.bulk_ec_dc_tc using a fitting approach
    - non_fitting : Calculate missing values of soil.df.bulk_ec_dc_tc using a non-fitting approach
    """
    if (np.isnan(soil.df.bulk_ec_dc_tc)).any():  # Go over if any value is missing      

        FrequencyEC(soil)
        Temperature(soil)
        shift_to_bulk_ec_dc_tc(soil)

        # Condition for fitting routine 
        if sum(not np.isnan(soil.water[x]) and not np.isnan(soil.df.bulk_ec_dc_tc[x]) for x in range(soil.n_states))>= 3:
            fitting(soil)

        # Condition for non-fitting routine 
        if any(not np.isnan(soil.df.water[x]) and np.isnan(soil.df.bulk_ec_dc_tc[x])  for x in range(soil.n_states)):
            non_fitting(soil)

    return soil.df.bulk_ec_dc_tc.values


def shift_to_bulk_ec_dc_tc(soil):
    """
    Compute missing values of soil.df.bulk_ec_dc_tc based on soil.df.bulk_ec or soil.df.bulk_ec_dc
    
    This function checks each soil state for the presence of non-NaN `bulk_ec` or `bulk_ec_dc` values, along with
    NaN `bulk_ec_dc_tc` values. When such a condition is met, it sequentially calls three functions to perform
    conversions and updates: `non_dc_to_dc` to convert non-direct current (non-DC) electrical conductivity (EC)
    values to direct current (DC) values, `non_dc_non_tc_to_dc_tc` to convert non-DC, non-temperature corrected (non-TC)
    EC values to DC, temperature-corrected (TC) values, and finally, `non_tc_to_tc` to convert non-temperature corrected
    EC values to temperature-corrected values.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            A DataFrame containing quantitative information of soil attributes for each state, including
            `bulk_ec`, `bulk_ec_dc`, and `bulk_ec_dc_tc`.
        - n_states: int
            The number of soil states.

    Returns
    -------
    None
        This function does not return a value. It operates in-place, modifying the soil object's DataFrame directly.

    External Functions
    ------------------
    non_dc_to_dc : Calculate missing values of soil.df.bulk_ec_dc based on soil.df.bulk_ec
    non_dc_non_tc_to_dc_tc : Calculate missing values of soil.df.bulk_ec_dc_tc based on soil.df.bulk_ec
    non_tc_to_tc : Calculate missing values of soil.df.bulk_ec_dc_tc based on soil.df.bulk_ec_dc

    Notes
    -----
    - The function operates in-place, modifying the `soil.df` DataFrame directly.
    - It is designed to handle soil states that have either `bulk_ec` or `bulk_ec_dc` values available but are missing `bulk_ec_dc_tc` values.
    - This sequential update process ensures that all relevant EC values are converted and updated to their DC, TC equivalents where necessary.

    """    
    if any(((not np.isnan(soil.df.bulk_ec[x])) or (not np.isnan(soil.df.bulk_ec_dc[x]))) and np.isnan(soil.df.bulk_ec_dc_tc[x]) for x in range(soil.n_states)):
        non_dc_to_dc(soil) 
        non_dc_non_tc_to_dc_tc(soil) 
        non_tc_to_tc(soil)


def non_dc_non_tc_to_dc_tc(soil):
    """
    Calculate missing values of soil.df.bulk_ec_dc_tc based on soil.df.bulk_ec

    If a `bulk_ec_dc_tc` value is missing (`NaN`) and the corresponding soil temperature is 298.15K with an electrical conductivity (EC) measurement frequency of 5Hz or less, 
    the `bulk_ec_dc_tc` value is set equal to the non-DC, non-temperature corrected `bulk_ec` value for that state. 
    Additionally, an annotation is added to `soil.info['bulk_ec_dc_tc']` indicating that the updated value is equal to the `bulk_ec` value under the specified conditions, 
    as part of the `predict.bulk_ec_dc_tc.non_dc_non_tc_to_dc_tc` process.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            A DataFrame containing quantitative information of soil attributes for each state, including
            `bulk_ec`, `bulk_ec_dc_tc`, `temperature`, and `frequency_ec`.
        - info: DataFrame
            Data Frame containing descriptive information about how each array-like attribute was calculated.
        - n_states: int
            The number of soil states.

    Returns
    -------
    None
        This function does not return a value. It operates in-place, modifying the soil object's DataFrame and info dictionary directly.

    Notes
    -----
    - The function is specifically designed to handle cases where direct measurements of temperature-corrected DC bulk EC are not available but can be inferred from non-DC EC measurements under standard conditions (298.15K and ≤5Hz).
    - Annotations in `soil.info['bulk_ec_dc_tc']` aid in tracking the source and method used for estimating the updated `bulk_ec_dc_tc` values.
    """
    missing_bulk_ec_dc_tc_before = soil.df['bulk_ec_dc_tc'].isna() 

    soil.df['bulk_ec_dc_tc'] = [soil.df.bulk_ec[x] 
                                if np.isnan(soil.df.bulk_ec_dc_tc[x]) and soil.df.temperature[x] == 298.15 and soil.df.frequency_ec[x] <= 5 
                                else soil.df.bulk_ec_dc_tc[x] for x in range(soil.n_states)]
    
    missing_bulk_ec_dc_tc_after = soil.df['bulk_ec_dc_tc'].isna()
    
    soil.info['bulk_ec_dc_tc'] = [str(soil.info.bulk_ec_dc_tc[x]) + (
            "--> Equal to soil.df.bulk_ec in predict.bulk_ec_dc_tc.non_dc_non_tc_to_dc_tc"
            if missing_bulk_ec_dc_tc_before[x] and not missing_bulk_ec_dc_tc_after[x]
            else "--> Provide bulk_ec_dc_tc; otherwise, bulk_ec, temperature, and frequency_ec"
            if missing_bulk_ec_dc_tc_before[x] and missing_bulk_ec_dc_tc_after[x]
            else "")
        if missing_bulk_ec_dc_tc_before[x]
        else soil.info.bulk_ec_dc_tc[x]
        for x in range(soil.n_states)]
    
    
def non_tc_to_tc(soil):
    """
    Calculate missing values of soil.df.bulk_ec_dc_tc based on soil.df.bulk_ec_dc

    This function iterates over the soil states to update `bulk_ec_dc_tc` values. 
    For states where `bulk_ec_dc_tc` is missing and the temperature is exactly 298.15K, the value is set equal to the non-temperature corrected `bulk_ec_dc`. 
    An annotation is added to `soil.info['bulk_ec_dc_tc']` indicating that the updated value is equivalent to `soil.df.bulk_ec_dc` under standard temperature conditions, 
    as part of the `predict.bulk_ec_dc_tc.non_tc_to_tc` process.
    If the temperature deviates from 298.15K and `bulk_ec_dc_tc` is missing, the `SheetsHendrickx` function is used to calculate the temperature-corrected value from `bulk_ec_dc`. 
    This is also annotated in `soil.info['bulk_ec_dc_tc']` to indicate that the value was calculated using the `SheetsHendrickx` function under non-standard temperature conditions.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            A DataFrame containing quantitative information of soil attributes for each state, including
            `bulk_ec_dc`, `bulk_ec_dc_tc`, and `temperature`.
        - info: DataFrame
            A dictionary containing descriptive information about how each array-like attribute was calculated.
        - n_states: int
            The number of soil states.

    Returns
    -------
    None
        This function does not return a value. It operates in-place, modifying the soil object's DataFrame and info dictionary directly.

    External Functions
    ------------------
    SheetsHendrickx : Calculate the soil bulk real electrical conductivity using the Sheets-Hendricks model and return

    Notes
    -----
    - The function distinguishes between standard (298.15K) and non-standard temperatures for updating `bulk_ec_dc_tc` values.
    - Annotations in `soil.info['bulk_ec_dc_tc']` provide insight into the source of the updated values, enhancing data traceability.
    - The `SheetsHendrickx` function is utilized for temperature corrections under non-standard conditions, applying a model to estimate the temperature-corrected EC value.    
    """ 
    missing_bulk_ec_dc_tc_before = soil.df['bulk_ec_dc_tc'].isna() 

    soil.df['bulk_ec_dc_tc'] = [soil.df.bulk_ec_dc[x] if np.isnan(soil.df.bulk_ec_dc_tc[x]) and soil.df.temperature[x] == 298.15 else soil.df.bulk_ec_dc_tc[x] for x in range(soil.n_states)]

    missing_bulk_ec_dc_tc_after = soil.df['bulk_ec_dc_tc'].isna()

    soil.info['bulk_ec_dc_tc'] = [str(soil.info.bulk_ec_dc_tc[x]) + (
            "--> Equal to soil.df.bulk_ec_dc in predict.bulk_ec_dc_tc.non_tc_to_tc"
            if missing_bulk_ec_dc_tc_before[x] and not missing_bulk_ec_dc_tc_after[x]
            else "--> Provide bulk_ec_dc_tc; otherwise, bulk_ec_dc and temperature"
            if missing_bulk_ec_dc_tc_before[x] and missing_bulk_ec_dc_tc_after[x]
            else "")
        if missing_bulk_ec_dc_tc_before[x]
        else soil.info.bulk_ec_dc_tc[x]
        for x in range(soil.n_states)]
    

    missing_bulk_ec_dc_tc_before = soil.df['bulk_ec_dc_tc'].isna() 

    soil.df['bulk_ec_dc_tc'] = [SheetsHendrickx(soil.df.bulk_ec_dc[x], soil.df.temperature[x]) 
                                if np.isnan(soil.df.bulk_ec_dc_tc[x]) and soil.df.temperature[x] != 298.15 
                                else soil.df.bulk_ec_dc_tc[x] 
                                for x in range(soil.n_states)]
    
    missing_bulk_ec_dc_tc_after = soil.df['bulk_ec_dc_tc'].isna()

    soil.info['bulk_ec_dc_tc'] = [str(soil.info.bulk_ec_dc_tc[x]) + (
            "--> Calculated using SheetsHendrickx function in predict.bulk_ec_dc_tc.non_tc_to_tc"
            if missing_bulk_ec_dc_tc_before[x] and not missing_bulk_ec_dc_tc_after[x]
            else "--> Provide bulk_ec_dc_tc; otherwise, bulk_ec_dc and temperature"
            if missing_bulk_ec_dc_tc_before[x] and missing_bulk_ec_dc_tc_after[x]
            else "")
        if missing_bulk_ec_dc_tc_before[x]
        else soil.info.bulk_ec_dc_tc[x]
        for x in range(soil.n_states)]
    

def fitting(soil):
    """ 
    Calculate missing values of soil.df.bulk_ec_dc_tc using a fitting approach

    This function utilizes the WunderlichEC model to estimate the soil's bulk real electrical conductivity at DC frequency temperature-corrected based on water content. 
    It calculates the model's parameters and fits them to the provided calibration data. 
    The accuracy of the fitting is determined by the R2 score. 

    Parameters
    ----------
    soil : object
        A custom soil object that contains:

        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state.
            Includes: water, water_ec, and bulk_ec_dc_tc
        - info : DataFrame
            Data Frame containing descriptive information about how each array-like attribute was determined or modified.
        - Lw : float
            Soil scalar depolarization factor of water aggregates (effective medium theory)
        - roundn : int
            Number of decimal places to round results.
        - range_ratio : float
            Ratio to extend the domain of the regression by fitting approach.
        - n_states : int
            Number of soil states. 

    Returns
    -------
    None 
        This function does not return a value. It operates in-place, modifying the soil object's DataFrame and info dictionary directly.

    Notes
    -----
    This function modifies the soil object in-place by updating the `df` and `info` dataframes.
    The function either estimates or uses the known Lw parameter for the WunderlichEC model and 
    fits the model to the calibration data.

    External Functions
    ------------------
    WunderlichEC: Calculate the soil bulk real electrical conductivity using the Wunderlich model and return
    WaterEC: Calculate missing values of soil.df.water based on soil.df.bulk_ec_dc_tc 
    """
    from .water_ec import WaterEC # Lazy import to avoid circular dependency

    WaterEC(soil)                    

    # Defining model parameters
    valids = ~np.isnan(soil.df.water) & ~np.isnan(soil.df.bulk_ec_dc_tc) # States where calibration data are
    water_init = min(soil.df.water[valids])
    bulk_ec_dc_tc_init = min(soil.df.bulk_ec_dc_tc[valids])
    water_final = max(soil.df.water[valids])
    water_range = [round(water_init - (water_final-water_init)/soil.range_ratio, soil.roundn), 
                  round(water_final + (water_final-water_init)/soil.range_ratio, soil.roundn)]
    if water_range[0] < 0:
        water_range[0] = 0
        
    # Obtain Lw attribute if unknown
    if np.isnan(soil.Lw):

        # Defining minimization function to obtain Lw
        def objective_Lw(Lw):
            wund_eval = [WunderlichEC(soil.df.water[x], bulk_ec_dc_tc_init, water_init, soil.df.water_ec[x], Lw)[0] if valids[x] else np.nan for x in range(soil.n_states)]    
            Lw_RMSE = np.sqrt(np.nanmean((np.array(wund_eval) - soil.df.bulk_ec_dc_tc)**2))
            return Lw_RMSE
    
        # Calculating optimal Lw
        result = minimize(objective_Lw, 0.1, bounds=[(-0.2, 0.8)], method='L-BFGS-B')
        soil.Lw = result.x[0]

    # If Lw is known
    if ~np.isnan(soil.Lw):
        if not isinstance(soil.Lw, np.floating):
            soil.Lw = soil.Lw[0]
        # Calculating the R2 score of the model fitting
        R2 = round(R2_score(soil.df.bulk_ec_dc_tc, WunderlichEC(soil.df.water, bulk_ec_dc_tc_init, water_init, soil.df.water_ec, soil.Lw)), soil.roundn)

        missing_bulk_ec_dc_tc_before = soil.df['bulk_ec_dc_tc'].isna() 

        soil.df['bulk_ec_dc_tc'] = [round(WunderlichEC(soil.df.water[x], bulk_ec_dc_tc_init, water_init, soil.df.water_ec[x], soil.Lw), soil.roundn+3) 
                                    if np.isnan(soil.df.bulk_ec_dc_tc[x]) and (min(water_range) <= soil.water[x] <= max(water_range)) 
                                    else soil.df.bulk_ec_dc_tc[x] for x in range(soil.n_states)]
        
        missing_bulk_ec_dc_tc_after = soil.df['bulk_ec_dc_tc'].isna() 

        # Update info for calculated bulk_ec_dc_tc
        soil.info['bulk_ec_dc_tc'] = [str(soil.info.bulk_ec_dc_tc[x]) + (
                "--> Calculated by fitting (R2="+str(R2)+") WunderlichEC function in predict.bulk_ec_dc_tc.fitting, for soil.water values between"+str(water_range)
                if missing_bulk_ec_dc_tc_before[x] and not missing_bulk_ec_dc_tc_after[x]
                else "--> Provide bulk_ec_dc_tc; otherwise, water and water_ec. Regression valid for water values between"+str(water_range)
                if missing_bulk_ec_dc_tc_before[x] and missing_bulk_ec_dc_tc_after[x]
                else "")
            if missing_bulk_ec_dc_tc_before[x]
            else soil.info.bulk_ec_dc_tc[x]
            for x in range(soil.n_states)]
        

def non_fitting(soil):
    """ 
    Calculate missing values of soil.df.bulk_ec_dc_tc using a non-fitting approach

    This function employs the Fu function (reported with an R^2 of 0.98) to estimate the 
    soil's bulk real electrical conductivity at DC frequency temperature-corrected based on volumetric water content.

    Parameters
    ----------
    soil : object
        A custom soil object that contains:

        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state.
            Includes: water, clay, porosity, bulk_ec, water_ec, solid_ec, dry_ec, sat_ec, and bulk_ec_dc_tc.
        - info : DataFrame
            Data Frame containing descriptive information about how each array-like attribute was determined or modified.
        - roundn : int
            Number of decimal places to round results.
        - n_states : int
            Number of soil states.

    Returns
    -------
    None
        This function does not return a value. It operates in-place, modifying the soil object's DataFrame and info dictionary directly.

    External functions
    --------
    Fu: Calculate the soil bulk real electrical conductivity using the Fu model and return
    Texture: Calculate missing values of soil.df.sand, soil.df.silt, and soil.df.clay and return
    Porosity: Calculate missing values of soil.df.porosity and return
    WaterEC: Compute missing values of soil.df.water_ec and return  
    SolidEC: Set missing values of soil.df.solid_ec and return
    
    Notes
    -----
    This function modifies the soil object in-place by updating the `df` and `info` dataframes.
    The function uses optimization techniques to minimize the difference between the Fu function output 
    and the provided bulk real DC electrical conductivity temperature-corrected to determine the volumetric water content.
    """
    from .water_ec import WaterEC # Lazy import to avoid circular dependency

    Texture(soil)
    Porosity(soil)
    WaterEC(soil)
    SolidEC(soil)
 
    missing_bulk_ec_dc_tc_before = soil.df['bulk_ec_dc_tc'].isna() 

    soil.df['bulk_ec_dc_tc'] = [round(Fu(soil.df.water[x], soil.df.clay[x], soil.df.porosity[x], soil.df.water_ec[x], soil.df.solid_ec[x], soil.df.dry_ec[x], soil.df.sat_ec[x]), soil.roundn+3) 
                             if np.isnan(soil.df.bulk_ec_dc_tc[x]) 
                             else soil.df.bulk_ec_dc_tc[x] for x in range(soil.n_states)]
    
    missing_bulk_ec_dc_tc_after = soil.df['bulk_ec_dc_tc'].isna()  

    soil.info['bulk_ec_dc_tc'] = [str(soil.info.bulk_ec_dc_tc[x]) + (
            "--> Calculated using Fu function (reported R2=0.98) in predict.bulk_ec_dc_tc.non_fitting"
            if missing_bulk_ec_dc_tc_before[x] and not missing_bulk_ec_dc_tc_after[x]
            else "--> Provide bulk_ec_dc_tc; otherwise, water, clay, porosity, and water_ec; or dry_ec, sat_ec, porosity, clay, and water; or dry_ec, water_ec, porosity, clay, and water"
            if missing_bulk_ec_dc_tc_before[x] and missing_bulk_ec_dc_tc_after[x]
            else "")
        if missing_bulk_ec_dc_tc_before[x]
        else soil.info.bulk_ec_dc_tc[x]
        for x in range(soil.n_states)]




import numpy as np
from scipy.optimize import minimize

from pedophysics.pedophysical_models.bulk_perm import WunderlichP, LongmireSmithP, LR, LR_W, LR_MV
from pedophysics.utils.stats import R2_score

from .water_perm import *
from .frequency_perm import *
from .bulk_ec import *
from .bulk_perm_inf import *
from .porosity import *
from .air_perm import *
from .solid_perm import *
from .temperature import *
from .texture import *

def BulkPerm(soil):
    """ 
    Computes missing values of soil.df.bulk_perm and return.

    Uses the given electromagnetic frequency to determine the method to estimate bulk permittivity. 
    If no frequency is provided or all frequencies are the same, different methods or conditions 
    are invoked.

    Parameters
    ----------
    soil : object
        A custom soil object containing:

        - df : DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state.
            Includes: bulk_perm and frequency_perm.
        - n_states : int
            Number of soil states
        - info : DataFrame
            Data Frame containing descriptive information about how each array-like attribute was determined or modified.

    Returns
    -------
    numpy.ndarray
        soil.df.bulk_perm.values: Array containing updated soil bulk real relative dielectric permittivity values

    Notes
    -----
    The function modifies the soil object in-place by updating the `df` and `info` attributes based 
    on the given conditions and returns the updated values of bulk permittivity.

    External Functions
    ------------------
    - FrequencyPerm : Set missing values of soil.df.frequency_perm and return 
    - fixed_freq, changing_freq : Functions invoked based on the condition of the EM frequency provided.

    Example
    -------
    >>> sample = Soil( water = [0.3, 0.1, 0.15, 0.23, 0.02],
                porosity = 0.434,
                texture = 'Silt loam',
                instrument = 'GPR')

    >>> predict.BulkPerm(sample) 
    array([18.563,  6.781,  9.328, 13.973,  3.357])
    """

    if (np.isnan(soil.df.bulk_perm)).any():  # Go over if any value is missing        
        FrequencyPerm(soil)

        # Condition to ask for frequency data
        if (np.isnan(soil.df.frequency_perm)).all():
            soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + "--> Provide  frequency_perm" for x in range(soil.n_states)]

        # Condition for fixed EM frequency
        elif np.all(soil.df.frequency_perm == soil.df.frequency_perm[0]):
            fixed_freq(soil)

        # Condition for changing EM frequency
        else:
            changing_freq(soil)

    return soil.df.bulk_perm.values


def fixed_freq(soil):
    """ 
    Decide between fitting and non-fitting approaches.

    Based on the given soil state data, chooses between two methodologies - fitting or non-fitting 
    to determine the bulk permittivity values for a fixed frequency.

    Parameters
    ----------
    soil : object
        A custom soil object containing:

        - df : DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state.
            Includes: bulk_perm, water, and frequency_perm.
        - n_states : int
            Number of soil states

    Returns
    -------
    None
        
    Notes
    -----
    The function modifies the soil object in-place based on the given conditions.

    External Functions
    ------------------
    - fitting : Calculate missing values of soil.df.bulk_perm using a fitting approach
    - non_fitting : Calculate missing values of soil.df.bulk_perm using a non fitting approach.
    """

    # Condition for fitting approach
    if sum(not np.isnan(soil.water[x]) and not np.isnan(soil.bulk_perm[x]) for x in range(soil.n_states)) >= 3 :
        fitting(soil)

    # Condition for non-fitting approach
    if any(np.isnan(soil.df.bulk_perm[x]) and not np.isnan(soil.df.water[x]) and soil.df.frequency_perm[x] >= 5 and soil.df.frequency_perm[x] <= 30e9 for x in range(soil.n_states) ):
        non_fitting(soil)

        
def fitting(soil):
    """ 
    Calculate missing values of soil.df.bulk_perm using a fitting approach

    This function fits the WunderlichP function to the soil data to determine the bulk permittivity values.
    It can either calculate the Lw parameter or use a provided value. The accuracy of the fitting is 
    determined by the R2 score. 

    Parameters
    ----------
    soil : object
        A custom soil object containing:

        - df : DataFrame
            Data Frame containing all the quantitative information of soil array-like attributes for each state.
            Includes: water, water_perm, and bulk_perm
        - info : DataFrame
            Data Frame containing descriptive information about how each array-like attribute was determined or modified.
        - Lw : float
            Soil scalar depolarization factor of water aggregates (effective medium theory)
        - roundn : int
            Number of decimal places to round results.
        - range_ratio : float
            Ratio to extend the domain of the regression by fitting approach.
        - n_states : int
            Number of soil states

    Returns
    -------
    None
        The function directly modifies the `soil` object's `df` and `info` attributes and does not return any value.

    Notes
    -----
    The function modifies the soil object in-place by updating the `df` and `info` attributes based on the fitting results.
    
    External Functions
    ------------------
    - Temperature : Set missing values of soil.df.temperature and return 
    - WaterPerm : Calculate or set missing values of soil.df.water_perm and return
    - WunderlichP : Calculate the soil bulk real relative dielectric permittivity using the Wunderlich model and return
    """

    Temperature(soil)
    WaterPerm(soil)                      

    # Defining model parameters
    valids = ~np.isnan(soil.df.water) & ~np.isnan(soil.df.bulk_perm) # States where calibration data are
    water_init = min(soil.df.water[valids])
    bulk_perm_init = min(soil.df.bulk_perm[valids])
    water_final = max(soil.df.water[valids])
    water_range = [round(water_init - (water_final-water_init)/soil.range_ratio, soil.roundn), 
                  round(water_final + (water_final-water_init)/soil.range_ratio, soil.roundn)]
    if water_range[0] < 0:
        water_range[0] = 0

    # Obtain Lw attribute if unknown
    if np.isnan(soil.Lw):

        # Defining minimization function to obtain Lw
        def objective_Lw(Lw):
            wund_eval = [WunderlichP(soil.df.water[x], bulk_perm_init, water_init, soil.df.water_perm[x], Lw)[0] if valids[x] else np.nan for x in range(soil.n_states)]    
            Lw_RMSE = np.sqrt(np.nanmean((np.array(wund_eval) - soil.df.bulk_perm)**2))
            return Lw_RMSE
    
        # Calculating optimal Lw
        result = minimize(objective_Lw, 0.1, bounds=[(-0.2, 0.8)], method='L-BFGS-B')
        soil.Lw = result.x[0]

    # If Lw is known
    if ~np.isnan(soil.Lw):
        if not isinstance(soil.Lw, np.floating):
            soil.Lw = soil.Lw[0]
        # Calculating the R2 score of the model fitting
        R2 = round(R2_score(soil.df.bulk_perm, WunderlichP(soil.df.water, bulk_perm_init, water_init, soil.df.water_perm, soil.Lw)), soil.roundn)

        # Check for missing values
        missing_bulk_perm_before = soil.df['bulk_perm'].isna()

        soil.df['bulk_perm'] = [round(WunderlichP(soil.df.water[x], bulk_perm_init, water_init, soil.df.water_perm[x], soil.Lw), soil.roundn) 
                              if (min(water_range) <= soil.water[x] <= max(water_range)) and np.isnan(soil.df.bulk_perm[x]) 
                              else soil.df.bulk_perm[x] for x in range(soil.n_states)]

        missing_bulk_perm_after = soil.df['bulk_perm'].isna()

        # Saving calculated bulk_perm and its info with R2 and valid water range
        soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + (
                "--> Calculated by fitting (R2="+str(R2)+") WunderlichP function in predict.bulk_perm.fitting, for water values between"+str(water_range)
                if missing_bulk_perm_before[x] and not missing_bulk_perm_after[x]
                else "--> Provide bulk_perm; otherwise, water. Regression valid for water values between"+str(water_range)
                if missing_bulk_perm_before[x] and missing_bulk_perm_after[x]
                else "")
            if missing_bulk_perm_before[x]
            else soil.info.bulk_perm[x]
            for x in range(soil.n_states)]


def non_fitting(soil):
    """ 
    Calculate missing values of soil.df.bulk_perm using a non fitting approach

    This function determines the bulk permittivity of soil based on given conditions and known empirical relationships. 
    Depending on the electromagnetic (EM) frequency range provided, various functions are used.

    Parameters
    ----------
    soil : object
        A custom soil object containing:

        - df : DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state.
            Includes: water, bulk_perm, frequency_perm, bulk_ec, bulk_perm_inf, porosity, clay, air_perm, solid_perm, water_perm, and CEC.
        - roundn : int
            Number of decimal places to round results.

    Returns
    -------
    None
        The function directly modifies the `soil` object's `df` and `info` attributes and does not return any value.

    Notes
    -----
    The function modifies the soil object in-place by updating the `df` and `info` attributes based on the chosen approach.
    
    External Functions
    ------------------
    - BulkPermInf : Determines the bulk permittivity at infinite frequency.
    - BulkEC : Estimates the soil's bulk electrical conductivity.
    - Temperature : Set missing values of soil.df.temperature and return 
    - Porosity : Calculate missing values of soil.df.porosity and return
    - AirPerm : Set missing values of soil.df.air_perm and return
    - SolidPerm : Set missing values of soil.df.solid_perm and return
    - WaterPerm : Calculate or set missing values of soil.df.water_perm and return
    - Texture : Calculate missing values of soil.df.sand, soil.df.silt, and soil.df.clay and return
    - LongmireSmithP : Calculate the soil bulk real relative dielectric permittivity using the Longmire-Smith model and return
    - LR_MV : Calculate the soil volumetric water content using the Lichtenecker and Rother model modified by Mendoza-Veirana and return
    - LR : Calculate the soil volumetric water content using the Lichtenecker and Rother model.
    - LR_W : Calculate the soil volumetric water content using the Lichtenecker and Rother model modified by Wunderlich and return
    """

    # Condition for lowest EM frequency
    if any(np.isnan(soil.df.bulk_perm[x]) and soil.df.frequency_perm[x] >= 5 and soil.df.frequency_perm[x] < 30e6 for x in range(soil.n_states)):

        BulkPermInf(soil)              
        BulkECDC(soil)
        
        missing_bulk_perm_before = soil.df['bulk_perm'].isna()

        soil.df['bulk_perm'] = [round(LongmireSmithP(soil.df.bulk_ec_dc[x], soil.df.bulk_perm_inf[x], soil.df.frequency_perm[x]), soil.roundn) 
                                if np.isnan(soil.df.bulk_perm[x]) 
                                else soil.df.bulk_perm[x] for x in range(soil.n_states)]

        missing_bulk_perm_after = soil.df['bulk_perm'].isna()

        # Saving calculated bulk_perm and its info        
        soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + (
                "--> Calculated using LongmireSmithP function in predict.bulk_perm.non_fitting" 
                if missing_bulk_perm_before[x] and not missing_bulk_perm_after[x]
                else "--> Provide bulk_perm; otherwise, bulk_ec_dc and frequency_perm" 
                if missing_bulk_perm_before[x] and missing_bulk_perm_after[x]
                else "")
            if missing_bulk_perm_before[x]
            else soil.info.bulk_perm[x]
            for x in range(soil.n_states)]
        
    # Condition for EM frequency of common moisture sensors and GPR
    elif (np.isnan(soil.df.bulk_perm)).any() & ((soil.df.frequency_perm >= 30e6) & (soil.df.frequency_perm <= 30e9)).all(): 
        Temperature(soil)
        Porosity(soil)                      
        AirPerm(soil)                   
        SolidPerm(soil)                  
        WaterPerm(soil)               
        Texture(soil)                    

        if ((soil.df.frequency_perm >= 30e6) & (soil.df.frequency_perm < 100e6)).all():
            
            missing_bulk_perm_before = soil.df['bulk_perm'].isna()
            soil.df['bulk_perm'] = [np.round(LR_MV(soil.df.water[x], soil.df.porosity[x], soil.df.air_perm[x], soil.df.solid_perm[x], soil.df.water_perm[x], soil.df.CEC[x]), soil.roundn) 
                                    if np.isnan(soil.df.bulk_perm[x]) 
                                    else soil.df.bulk_perm[x] for x in range(soil.n_states)]
            
            missing_bulk_perm_after = soil.df['bulk_perm'].isna()
            
            soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + (
                    "--> Calculated using LR_MV (reported R2=0.93) function in predict.bulk_perm.non_fitting" 
                    if missing_bulk_perm_before[x] and not missing_bulk_perm_after[x]
                    else "--> Provide bulk_perm; otherwise, water, porosity, and CEC" 
                    if missing_bulk_perm_before[x] and missing_bulk_perm_after[x]
                    else "")
                if missing_bulk_perm_before[x]
                else soil.info.bulk_perm[x]
                for x in range(soil.n_states)]
            

        elif ((soil.df.frequency_perm >= 100e6) & (soil.df.frequency_perm < 200e6)).all():

            if np.isnan(soil.alpha): soil.alpha = 0.5 
            
            missing_bulk_perm_before = soil.df['bulk_perm'].isna()
            soil.df['bulk_perm'] = [np.round(LR(soil.df.water[x], soil.df.porosity[x], soil.df.air_perm[x], soil.df.solid_perm[x], soil.df.water_perm[x], soil.alpha[x]), soil.roundn) 
                                    if np.isnan(soil.df.bulk_perm[x]) 
                                    else soil.df.bulk_perm[x] for x in range(soil.n_states)]
            missing_bulk_perm_after = soil.df['bulk_perm'].isna()
            
            soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + (
                    "--> Calculated using LR function (reported RMSE=0.032) in predict.bulk_perm.non_fitting" 
                    if missing_bulk_perm_before[x] and not missing_bulk_perm_after[x]
                    else "--> Provide bulk_perm; otherwise, water and porosity" 
                    if missing_bulk_perm_before[x] and missing_bulk_perm_after[x]
                    else "")
                if missing_bulk_perm_before[x]
                else soil.info.bulk_perm[x]
                for x in range(soil.n_states)]
            
        elif ((soil.df.frequency_perm >= 200e6) & (soil.df.frequency_perm <= 30e9)).all(): 
            
            missing_bulk_perm_before = soil.df['bulk_perm'].isna()

            soil.df['bulk_perm'] = [np.round(LR_W(soil.df.water[x], soil.df.porosity[x], soil.df.air_perm[x], soil.df.solid_perm[x], soil.df.water_perm[x], soil.df.clay[x]), soil.roundn) 
                                    if np.isnan(soil.df.bulk_perm[x]) 
                                    else soil.df.bulk_perm[x] for x in range(soil.n_states)] 
            
            missing_bulk_perm_after = soil.df['bulk_perm'].isna()
            
            soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + (
                    "--> Calculated using LR_W function in predict.bulk_perm.non_fitting" 
                    if missing_bulk_perm_before[x] and not missing_bulk_perm_after[x]
                    else "--> Provide bulk_perm; otherwise, water, porosity, and clay" 
                    if missing_bulk_perm_before[x] and missing_bulk_perm_after[x]
                    else "")
                if missing_bulk_perm_before[x]
                else soil.info.bulk_perm[x]
                for x in range(soil.n_states)]


def changing_freq(soil):
    """ 
    calculate missing values of soil.df.bulk_perm based on soil.df.bulk_ec_dc

    This function determines the bulk permittivity of soil based on given conditions and the LongmireSmithP pedophysical model. 
    It specifically addresses cases where the electromagnetic (EM) frequency is not constant.

    Parameters
    ----------
    soil : object
        A custom soil object containing:

        - df : DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state.
            Includes: bulk_perm, frequency_perm, bulk_ec_dc, clay, sand, and bulk_perm_inf.
        - roundn : int
            Number of decimal places to round results.

    Returns
    -------
    None
        The function directly modifies the `soil` object's `df` and `info` attributes and does not return any value.
        
    Notes
    -----
    The function modifies the soil object in-place by updating the `df` and `info` attributes.

    External Functions
    ------------------
    - BulkPermInf : Set missing values of soil.df.bulk_perm_inf and return
    - BulkECDC : Compute missing values of soil.df.bulk_ec_dc and return
    - LongmireSmithP : Calculate the soil bulk real relative dielectric permittivity using the Wunderlich model and return
    """
    Texture(soil)
    BulkPermInf(soil)             
    BulkECDC(soil)

    def warn_states(soil):
        # Warn about applying LongmireSmithP function to non-validated soil conditions
        mask_invalid = (
            (soil.df.frequency_perm > 200e6) &
            (soil.df.water > 0.22) &
            (soil.df.porosity > 0.255) &
            (soil.df.water_ec > 3.3) | (soil.df.water_ec < 0.0016) &
            (soil.df.clay > 10) | (soil.df.sand < 85) &
            np.isnan(soil.df.bulk_perm)
        )

        # Find the indices of the invalid states
        states_warns = np.where(mask_invalid)[0]

        if states_warns.size > 0:
            warnings.warn(
                f"LongmireSmithP function is applied to soil states {states_warns} with conditions "
                f"frequency_perm > 200e6, and water > 0.22, and porosity > 0.255, and water_ec > 3.3 or water_ec < 0.0016, and clay > 10 or sand < 85"
                f", for which the validity of such model is uncertain."
                )   

    warn_states(soil)
    missing_bulk_perm_before = soil.df['bulk_perm'].isna()
    
    soil.df['bulk_perm'] = [round(LongmireSmithP(soil.df.bulk_ec_dc[x], soil.df.bulk_perm_inf[x], soil.df.frequency_perm[x]), soil.roundn) 
                            if (np.isnan(soil.df.bulk_perm[x])) 
                            else soil.df.bulk_perm[x] 
                            for x in range(soil.n_states)]
    
    missing_bulk_perm_after = soil.df['bulk_perm'].isna()
    
    soil.info['bulk_perm'] = [str(soil.info.bulk_perm[x]) + (
            "--> Calculated using LongmireSmithP function in predict.bulk_perm.changing_freq" 
            if missing_bulk_perm_before[x] and not missing_bulk_perm_after[x]
            else "--> Provide bulk_perm; otherwise, bulk_ec_dc and frequency_perm" 
            if missing_bulk_perm_before[x] and missing_bulk_perm_after[x]
            else "")
        if missing_bulk_perm_before[x]
        else soil.info.bulk_perm[x]
        for x in range(soil.n_states)]


import numpy as np

def R2_score(actual, predicted):
    """
    Calculate the coefficient of determination (R^2) of a prediction.

    The R^2 score function computes the coefficient of determination, often used 
    to evaluate the performance of a regression model. The best possible score is 1.0.
    This function is designed to handle arrays with NaN values by ignoring such entries.

    Parameters
    ----------
    actual : array-like of shape (n_samples,)
        Ground truth (correct) target values.
    predicted : array-like of shape (n_samples,)
        Estimated targets as returned by a classifier.

    Returns
    -------
    float
        R^2 of the prediction.

    Notes
    -----
    This function works with arrays that include NaN values, ignoring such entries 
    during the computation. Therefore, 'actual' and 'predicted' arrays can have 
    missing values, but they must be of the same shape.

    Example
    -------
    >>> actual = np.array([3, -0.5, 2, 7, 4.2])
    >>> predicted = np.array([2.5, 0.0, 2.1, 7.8, 5.3])
    >>> R2_score(actual, predicted)
    0.9228556485355649
    """
    valids = ~np.isnan(actual) & ~np.isnan(predicted)
    # Calculate the total sum of squares
    ss_tot = np.sum((actual[valids] - np.mean(actual[valids])) ** 2)
    # Calculate the residual sum of squares
    ss_res = np.sum((actual[valids] - predicted[valids]) ** 2)

    # Calculate the R2 score
    r2 = 1 - (ss_res / ss_tot)
    return r2


import numpy as np
from scipy.optimize import minimize

from pedophysics.utils.stats import R2_score
from pedophysics.pedophysical_models.bulk_ec import Fu, WunderlichEC

from .water_ec import WaterEC
from .porosity import Porosity
from .solid_ec import SolidEC
from .frequency_ec import FrequencyEC
from .texture import Texture


def WaterFromEC(soil):
    """ 
    Calculate missing values of soil.df.water based on soil.df.bulk_ec_dc_tc 

    This function evaluates the availability of water content and bulk electrical conductivity data (bulk_ec_dc_tc) across soil states. 
    A fitting approach is applied if there are at least three soil states with known water content and bulk electrical conductivity. 
    A non-fitting approach is considered when water content is unknown and bulk electrical conductivity is known for any soil state.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state. 
            Includes: `water` and `bulk_ec_dc_tc`.
        - info: DataFrame
            Data Frame containing descriptive information about how each array-like attribute was calculated.
        - n_states: int
            The number of soil states represented in the `df`.

    Returns
    -------
    None

    Notes
    -----
    - The fitting approach requires at least three soil states with known water content and bulk electrical conductivity for reliable estimation.
    - The non-fitting approach is applied to individual soil states where water content is unknown but bulk electrical conductivity is available.

    External Functions
    ------------------
    FrequencyEC : Set missing values of soil.df.frequency_ec and return 
    fitting : Calculate missing values of soil.df.water using a fitting approach.
    non_fitting : Calculate missing values of soil.df.water using a non-fitting approach.

    Example
    -------
    >>> sample = Soil( bulk_ec = [0.01, np.nan, 0.025, 0.030, 0.040],
                clay = 10,
                porosity = 0.47,
                water_ec = 0.5)

    >>> WaterFromEC(sample) 
    >>> sample.df.water
    0    0.105
    1    Nan
    2    0.185
    3    0.206
    4    0.243
    Name: water, dtype: float64
    """
    FrequencyEC(soil)

    # Check for conditions to use a fitting approach
    if sum(not np.isnan(soil.water[x]) and not np.isnan(soil.df.bulk_ec_dc_tc[x]) for x in range(soil.n_states)) >= 3:
        fitting(soil)

    # Check for conditions to use a non-fitting approach
    if any(np.isnan(soil.df.water[x]) and not np.isnan(soil.df.bulk_ec_dc_tc[x]) for x in range(soil.n_states)):
        non_fitting(soil)


def non_fitting(soil):
    """ 
    Calculate missing values of soil.df.water using a non-fitting approach.


    This function applies the Fu function within a minimization process to estimate soil water content based on soil properties such as 
    clay content, porosity, water electrical conductivity (EC), solid EC, dry EC, and saturated EC. 
    The estimation is performed for each soil state where water content is unknown.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state. 
            Includes: `clay`, `porosity`, `water_ec`, `solid_ec`, `dry_ec`, `sat_ec`, `bulk_ec_dc_tc`, and potentially `water`.
        - info: DataFrame
            Data Frame containing descriptive information about how each array-like attribute was calculated.
        - n_states: int
            The number of soil states represented in the `df`.
        - roundn: int
            The number of decimal places for rounding estimated water content values.

    bulk_ec_dc : array-like
        Soil bulk real electrical conductivity at DC frequency [S/m].

    Notes
    -----
    - The Fu function is utilized in a minimization process to estimate water content by minimizing the difference between the estimated and actual bulk ECDCTC.
    - The estimation process is applied to each soil state where water content is unknown.


    External functions
    --------
    Fu: Calculate the soil bulk real electrical conductivity using the Fu model and return
    Texture: Calculate missing values of soil.df.sand, soil.df.silt, and soil.df.clay and return
    Porosity: Calculate missing values of soil.df.porosity and return
    WaterEC: Compute missing values of soil.df.water_ec and return  
    SolidEC: Set missing values of soil.df.solid_ec and return
    """    
    Texture(soil)
    Porosity(soil)
    WaterEC(soil)
    SolidEC(soil)

    # Defining minimization function to obtain water using Fu
    def objective_func_wat(x, clay, porosity, water_ec, solid_ec, dry_ec, sat_ec, EC):
        return (Fu(x, clay, porosity, water_ec, solid_ec, dry_ec, sat_ec) - EC)**2
    wat = []

    # Calculating water
    for i in range(soil.n_states):
        
        res = minimize(objective_func_wat, 0.15, args=(soil.df.clay[i], soil.df.porosity[i], soil.df.water_ec[i], soil.df.solid_ec[i], 
                                                        soil.df.dry_ec[i], soil.df.sat_ec[i], soil.df.bulk_ec_dc_tc[i]), bounds=[(0, .65)] )
        wat.append(np.nan if np.isnan(res.fun) else round(res.x[0], soil.roundn) )

    # Check for missing values
    missing_water_before = soil.df['water'].isna()

    soil.df['water'] = [round(wat[i], soil.roundn) if np.isnan(soil.df.water[i]) else soil.df.water[i] for i in range(soil.n_states) ]
    missing_water_after = soil.df['water'].isna()

    # Update info for calculated water
    soil.info['water'] = [str(soil.info.water[x]) + (
            "--> Calculated using Fu function (reported R2=0.98) in predict.water_from_ec.non_fitting"
            if missing_water_before[x] and not missing_water_after[x]
            else "--> Provide water; otherwise clay, porosity, water_ec and bulk_ec_dc_tc"
            if missing_water_before[x] and missing_water_after[x]
            else "")
        if missing_water_before[x]
        else soil.info.water[x]
        for x in range(soil.n_states)]


def fitting(soil):
    """ 
    Calculate missing values of soil.df.water using a fitting approach.

    This function evaluates soil states with known water content and bulk electrical conductivity to determine initial parameters for the WunderlichEC model. 
    If the Lw parameter associated with the model is unknown, it is optimized based on the root mean square error (RMSE) between estimated and actual bulk electrical conductivity. 
    Water content is then estimated for all soil states within a valid bulk electrical conductivity range using the optimized Lw parameter and the WunderlichEC model.

    Parameters
    ----------
    soil : Soil Object
        An object representing the soil, which must have the following attributes:
        - df: DataFrame
            Data Frame containing the quantitative information of all soil array-like attributes for each state. 
            Includes: `water`, `bulk_ec_dc_tc`, `water_ec`, and potentially `Lw`.
        - info: DataFrame
            Data Frame containing descriptive information about how each array-like attribute was calculated.
        - n_states: int
            The number of soil states represented in the `df`.
        - range_ratio: float
            A ratio used to determine the range of bulk electrical conductivity values considered valid for the model.
        - roundn: int
            The number of decimal places for rounding calculated water content values.
        - Lw: float or np.nan
            The WunderlichEC model parameter, if known; otherwise, np.nan.

    Returns
    -------
    None
        The function directly modifies the `soil` object's `df` and `info` attributes with the estimated water content and does not return any value.

    Notes
    -----
    This function modifies the soil object in-place by updating the `df` and `info` dataframes.
    The function either estimates or uses the known Lw parameter for the WunderlichEC model and 
    fits the model to the calibration data.

    External Functions
    ------------------
    WunderlichEC: Calculate the soil bulk real electrical conductivity using the Wunderlich model and return
    WaterEC: Compute missing values of soil.df.water_ec and return  
    """

    WaterEC(soil) 
    
    # Defining model parameters
    valids = ~np.isnan(soil.df.water) & ~np.isnan(soil.df.bulk_ec_dc_tc) # States where calibration data are
    water_init = np.nanmin(soil.df.water[valids])
    bulk_ec_init = np.nanmin(soil.df.bulk_ec_dc_tc[valids])
    bulk_ec_final = np.nanmax(soil.df.bulk_ec_dc_tc[valids])
    bulk_ec_range = [round(bulk_ec_init - (bulk_ec_final-bulk_ec_init)/soil.range_ratio, soil.roundn), 
                     round(bulk_ec_final + (bulk_ec_final-bulk_ec_init)/soil.range_ratio, soil.roundn)]
    if bulk_ec_range[0] < 0:
        bulk_ec_range[0] = 0

    # Obtain Lw attribute if unknown
    if np.isnan(soil.Lw):

        # Defining minimization function to obtain water
        def objective_Lw(Lw):
            wund_eval = [WunderlichEC(soil.df.water[x], bulk_ec_init, water_init, soil.df.water_ec[x], Lw)[0] if valids[x] else np.nan for x in range(soil.n_states)]    
            Lw_RMSE = np.sqrt(np.nanmean((np.array(wund_eval) - soil.df.bulk_ec_dc_tc)**2))
            return Lw_RMSE

        # Calculating optimal Lw
        result = minimize(objective_Lw, 0.1, bounds=[(-0.2, 0.8)], method='L-BFGS-B')
        soil.Lw = result.x[0]
        
    # If Lw is known
    if ~np.isnan(soil.Lw):
        if not isinstance(soil.Lw, np.floating):
            soil.Lw = soil.Lw[0]
        Wat_wund = []

        # Defining minimization function to obtain water
        def objective_wat(wat, i):
            Wat_RMSE = np.sqrt((WunderlichEC(wat, bulk_ec_init, water_init, soil.df.water_ec[i], soil.Lw) - soil.df.bulk_ec_dc_tc[i])**2)
            return Wat_RMSE
        
        # Looping over soil states to obtain water using WunderlichEC function
        for i in range(soil.n_states):
            if (min(bulk_ec_range) <= soil.df.bulk_ec_dc_tc[i] <= max(bulk_ec_range)) & ~np.isnan(soil.df.bulk_ec_dc_tc[i]):
                result = minimize(objective_wat, 0.15, args=(i), bounds=[(0, .65)], method='L-BFGS-B')
                Wat_wund.append(np.nan if np.isnan(result.fun) else round(result.x[0], soil.roundn))

            else:
                Wat_wund.append(np.nan)

        # Calculating the R2 score of the model fitting
        R2 = round(R2_score(soil.df.water[valids], np.array(Wat_wund)[valids]), soil.roundn)
    
        missing_water_before = soil.df['water'].isna()  

        soil.df['water'] = [Wat_wund[x] if np.isnan(soil.df.water[x]) else soil.df.water[x] for x in range(soil.n_states)]

        missing_water_after = soil.df['water'].isna()  

        soil.info['water'] = [str(soil.info.water[x]) + (
                "--> Calculated by fitting (R2="+str(R2)+") WunderlichEC function in predict.water_from_ec.fitting, for soil.bulk_ec values between: "+str(bulk_ec_range)
                if missing_water_before[x] and not missing_water_after[x]
                else "--> Provide water; otherwise, bulk_ec_dc_tc and water_ec. Regression valid for bulk_ec_dc_tc values between: "+str(bulk_ec_range)
                if missing_water_before[x] and missing_water_after[x]
                else "")
            if missing_water_before[x]
            else soil.info.water[x]
            for x in range(soil.n_states)]
    
        
